<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
 <head>
 <title>&lt;graph&gt;-Plugin for Mediawiki - Syntax</title>
 <meta name="MSSmartTagsPreventParsing" content="TRUE">
 <meta http-equiv="imagetoolbar" content="no">
 <link rel="stylesheet" type="text/css" href="base.css">
 <style type="text/css">
<!--
 .graph { margin-left: 2em; }
 pre { margin-right: 1.5em; }
 h3 { border-bottom: 1px solid #404040; padding-bottom: 0.3em; }
 a.top { top: -3.5em; }
-->
 </style>
</head>
<body bgcolor=white text=black>

<a name="top"></a>

<div class="menu">
  <p class="menu"><a class="menu" href="index.html" title="Back to the main page">Main</a></p>
</div>

<div class="right">

<h1>&lt;graph&gt;-Plugin for Mediawiki</h1>

<h2>Table of Contents:</h2>

<div class="text">
##TOC##

<p>
To see the input text for each graph, follow the <font color="red">Source</font> link in each section.
</p>

</div>

<h2>Fun with Graphs</h2>

<div class="text">
##HTML##
</div>

<div class="footer">

<p>
This page was automatically created at <strong><small>##time##</small></strong> by <code>examples/syntax.pl</code> running
<a href="http://search.cpan.org/~tels/Graph-Simple/" title="Get it from search.cpan.org">Graph::Easy</a> v##version##.
Contact <a href="/mail.html">Tels</a> for help.
</p>

</div>

</div> <!-- right cell ends here -->

</body></html>
